from tal.render import create, render


def create_nlos_scene(folder_name, args):
    create.create_nlos_scene(folder_name, args)


def render_nlos_scene(config_path, args):
    render.render_nlos_scene(config_path, args)
